from typing import Any, Union

from .store_file import store_file
from .load_file import load_file
from .store_file_local import store_file_local
from .TemporaryDirectory import TemporaryDirectory
from ._safe_pickle import _safe_pickle, _safe_unpickle

def store_text(text: str, *, label: Union[str, None]=None, cache_locally: bool=False, local: bool=False) -> str:
    with TemporaryDirectory() as tmpdir:
        fname = f'{tmpdir}/file.dat'
        with open(fname, 'w') as f:
            f.write(text)
        return store_file(fname, label=label, cache_locally=cache_locally, local=local)

def store_json(x: Any, *, separators=(',', ':'), indent=None, label: Union[str, None]=None, cache_locally: bool=False, local: bool=False) -> str:
    import simplejson
    text = simplejson.dumps(x, separators=separators, indent=indent, allow_nan=False, sort_keys=True)
    return store_text(text, label=label, cache_locally=cache_locally, local=local)

def store_npy(array: Any, *, label: Union[str, None]=None, cache_locally: bool=False, local: bool=False) -> str:
    import numpy as np
    with TemporaryDirectory() as tmpdir:
        fname = f'{tmpdir}/file.npy'
        np.save(fname, array, allow_pickle=False)
        return store_file(fname, label=label, cache_locally=cache_locally, local=local)

def store_pkl(x: Any, *, label: Union[str, None]=None, cache_locally: bool=False, local: bool=False) -> str:
    with TemporaryDirectory() as tmpdir:
        fname = f'{tmpdir}/file.npy'
        _safe_pickle(fname, x)
        return store_file(fname, label=label, cache_locally=cache_locally, local=local)

def load_text(uri: str, *, local_only: bool=False) -> Union[str, None]:
    local_path = load_file(uri, local_only=local_only)
    if local_path is None:
        return None
    with open(local_path, 'r') as f:
        return f.read()

def load_json(uri: str, *, local_only: bool=False) -> Union[dict, None]:
    import simplejson
    local_path = load_file(uri, local_only=local_only)
    if local_path is None:
        return None
    with open(local_path, 'r') as f:
        return simplejson.load(f)

def load_npy(uri: str, *, local_only: bool=False) -> Union[Any, None]:
    import numpy as np
    local_path = load_file(uri, local_only=local_only)
    if local_path is None:
        return None
    return np.load(local_path, allow_pickle=False)

def load_pkl(uri: str, *, local_only: bool=False) -> Union[Any, None]:
    local_path = load_file(uri, local_only=local_only)
    if local_path is None:
        return None
    return _safe_unpickle(local_path)

def store_text_local(text: str, label: Union[str, None]=None) -> str:
    with TemporaryDirectory() as tmpdir:
        fname = f'{tmpdir}/file.dat'
        with open(fname, 'w') as f:
            f.write(text)
        return store_file_local(fname, label=label)

def store_json_local(x: Any, *, separators=(',', ':'), indent=None, label: Union[str, None]=None) -> str:
    import simplejson
    text = simplejson.dumps(x, separators=separators, indent=indent, allow_nan=False, sort_keys=True)
    return store_text_local(text, label=label)

def store_npy_local(array: Any, label: Union[str, None]=None) -> str:
    import numpy as np
    with TemporaryDirectory() as tmpdir:
        fname = f'{tmpdir}/file.npy'
        np.save(fname, array, allow_pickle=False)
        return store_file_local(fname, label=label)

def store_pkl_local(x: Any, label: Union[str, None]=None) -> str:
    with TemporaryDirectory() as tmpdir:
        fname = f'{tmpdir}/file.npy'
        _safe_pickle(fname, x)
        return store_file_local(fname, label=label)


import os
import shutil
import requests
from typing import Union
from urllib.parse import quote
import time

from .get_kachery_cloud_dir import get_kachery_cloud_dir

from .load_file import _random_string
from ._fs_operations import _makedirs
from .store_file_local import _compute_file_hash, store_file_local
from ._custom_storage_backend import get_custom_storage_backend

def store_file(filename: str, *, label: Union[str, None] = None, cache_locally: bool = False, local: bool = False):
    if local:
        return store_file_local(filename, label=label)
    if os.environ.get('KACHERY_STORE_FILE_DIR') is not None:
        return store_file_local(filename, label=label, store_file_dir=os.environ['KACHERY_STORE_FILE_DIR'], store_file_prefix=os.getenv('KACHERY_STORE_FILE_PREFIX', None))

    _custom_storage_backend = get_custom_storage_backend()
    use_custom_storage_backend = _custom_storage_backend is not None and hasattr(_custom_storage_backend, 'store_file')
    if not use_custom_storage_backend:
        size = os.path.getsize(filename)
        alg = 'sha1'
        hash0 = _compute_file_hash(filename, algorithm=alg)
        uri = f'{alg}://{hash0}'
        kachery_zone = os.environ.get('KACHERY_ZONE', 'default')
        payload = {
            'type': 'initiateFileUpload',
            'size': size,
            'hashAlg': alg,
            'hash': hash0,
            'zone': kachery_zone
        }
        timer = time.time()
        while True:

            from ._kachery_gateway_request import _kachery_gateway_request
            response: dict = _kachery_gateway_request(payload)

            already_exists = response.get('alreadyExists', False)
            already_pending = response.get('alreadyPending', False)
            if already_exists:
                if label is not None:
                    uri = f'{uri}?label={quote(label)}'
                return uri
            elif already_pending:
                elapsed = time.time() - timer
                if elapsed > 60:
                    raise Exception(f'Upload is already pending... timeout: {uri}')
                print(f'Upload is already pending... waiting to retry {uri}')
                time.sleep(5)
            else:
                break

        signed_upload_url = response['signedUploadUrl']
        object_key = response['objectKey']
        with open(filename, 'rb') as f:
            resp_upload = requests.put(signed_upload_url, data=f)
            if resp_upload.status_code != 200:
                print(signed_upload_url)
                raise Exception(f'Error uploading file to bucket ({resp_upload.status_code}) {resp_upload.reason}: {resp_upload.text}')
        payload2 = {
            'type': 'finalizeFileUpload',
            'objectKey': object_key,
            'hashAlg': alg,
            'hash': hash0,
            'zone': kachery_zone
        }
        payload2['size'] = size
        from ._kachery_gateway_request import _kachery_gateway_request
        response2: dict = _kachery_gateway_request(payload2) # noqa

        if label is not None:
            uri = f'{uri}?label={quote(label)}'
    else:
        # custom storage backend
        assert _custom_storage_backend is not None
        uri = _custom_storage_backend.store_file(filename, label=label)
        hash0 = None # only computed if needed

    if cache_locally:
        kachery_cloud_dir = get_kachery_cloud_dir()
        if hash0 is None:
            # this would be None for custom storage backend
            hash0 = _compute_file_hash(filename, algorithm='sha1')
        e = hash0
        cache_parent_dir = f'{kachery_cloud_dir}/hash0/{e[0]}{e[1]}/{e[2]}{e[3]}/{e[4]}{e[5]}'
        if not os.path.exists(cache_parent_dir):
            _makedirs(cache_parent_dir)
        cache_filename = f'{cache_parent_dir}/{hash0}'
        if not os.path.exists(cache_filename):
            tmp_filename = f'{cache_filename}.tmp.{_random_string(8)}'
            shutil.copyfile(filename, tmp_filename)
            try:
                os.rename(tmp_filename, filename)
                # _chmod_file(filename)
            except: # noqa
                if not os.path.exists(cache_filename):
                    raise Exception(f'Problem renaming file: {tmp_filename} {filename}')

    return uri



import os

# _kachery_zone_info_by_zone = {}

_default_kachery_gateway_url = 'https://kachery-gateway.figurl.org'

# def _get_kachery_zone_info(kachery_zone: str):
#     if kachery_zone in _kachery_zone_info_by_zone:
#         return _kachery_zone_info_by_zone[kachery_zone]
#     from ._kachery_gateway_request import _kachery_gateway_request
#     resp = _kachery_gateway_request({
#         'type': 'getZoneInfo',
#         'zoneName': kachery_zone
#     }, kachery_gateway_url=_default_kachery_gateway_url) # important to use default here so we don't get an infinite recursion
#     _kachery_zone_info_by_zone[kachery_zone] = resp
#     return resp

def _get_kachery_gateway_url():
    url = os.environ.get('KACHERY_GATEWAY_URL', None)
    if url is not None:
        return url
    # kachery_zone = os.environ.get('KACHERY_ZONE', None)
    # if kachery_zone is not None:
    #     resp = _get_kachery_zone_info(kachery_zone)
    #     if resp['found']:
    #         return resp['kacheryGatewayUrl']
    #     else:
    #         raise Exception(f'Unrecognized kachery zone: {kachery_zone}')
    return _default_kachery_gateway_url